"""add task_run_org_run_id_index

Revision ID: b111f0f795bd
Revises: 60d0743274c9
Create Date: 2025-02-11 06:41:35.336836+00:00

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "b111f0f795bd"
down_revision: Union[str, None] = "60d0743274c9"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index("task_run_org_run_id_index", "task_runs", ["organization_id", "run_id"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("task_run_org_run_id_index", table_name="task_runs")
    # ### end Alembic commands ###
