# This file was auto-generated by Fern from our API Definition.

import typing

ProxyLocation = typing.Union[
    typing.Literal[
        "RESIDENTIAL",
        "US-CA",
        "US-NY",
        "US-TX",
        "US-FL",
        "US-WA",
        "RESIDENTIAL_ES",
        "RESIDENTIAL_IE",
        "RESIDENTIAL_GB",
        "RESIDENTIAL_IN",
        "RESIDENTIAL_JP",
        "RESIDENTIAL_FR",
        "RESIDENTIAL_DE",
        "RESIDENTIAL_NZ",
        "RESIDENTIAL_ZA",
        "RESIDENTIAL_AR",
        "RESIDENTIAL_ISP",
        "NONE",
    ],
    typing.Any,
]
