The user is trying to achieve a goal the web. Now they've decided to go through a list of values and take the same tasks with each variant in the list.

Help the user extract a list of values based on what they want to achieve:
```
{{ plan }}
```