"""Add workflow_run_id index to tasks

Revision ID: 12fb2dede685
Revises: a575628e1965
Create Date: 2024-10-09 16:52:36.095562+00:00

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "12fb2dede685"
down_revision: Union[str, None] = "a575628e1965"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f("ix_tasks_workflow_run_id"), "tasks", ["workflow_run_id"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_tasks_workflow_run_id"), table_name="tasks")
    # ### end Alembic commands ###
