"""add action_org_created_at_index

Revision ID: df80b5d155d0
Revises: 3aa0ef96942d
Create Date: 2025-01-29 07:38:58.641024+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "df80b5d155d0"
down_revision: Union[str, None] = "3aa0ef96942d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        "action_org_created_at_index", "actions", ["organization_id", sa.text("created_at DESC")], unique=False
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("action_org_created_at_index", table_name="actions")
    # ### end Alembic commands ###
