"""Add orgs.max_retries_per_step

Revision ID: ea8e24d0bc8e
Revises: 4630ab8c198e
Create Date: 2024-04-08 23:47:46.306300+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "ea8e24d0bc8e"
down_revision: Union[str, None] = "4630ab8c198e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("organizations", sa.Column("max_retries_per_step", sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("organizations", "max_retries_per_step")
    # ### end Alembic commands ###
