"""Add org_task_step_index

Revision ID: 68d78072fdb5
Revises: 24303f1669a7
Create Date: 2024-04-28 23:20:28.953686+00:00

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "68d78072fdb5"
down_revision: Union[str, None] = "24303f1669a7"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        "org_task_step_index",
        "artifacts",
        ["organization_id", "task_id", "step_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("org_task_step_index", table_name="artifacts")
    # ### end Alembic commands ###
