"""add max_steps_per_run to task

Revision ID: 8792454ce498
Revises: c4dca14a5e69
Create Date: 2024-05-11 21:04:38.384261+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "8792454ce498"
down_revision: Union[str, None] = "c4dca14a5e69"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("tasks", sa.Column("max_steps_per_run", sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("tasks", "max_steps_per_run")
    # ### end Alembic commands ###
