"""Add is_saved_task to workflows

Revision ID: 485667adef01
Revises: 2c163e606a3d
Create Date: 2024-06-27 19:49:41.506447+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "485667adef01"
down_revision: Union[str, None] = "2c163e606a3d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("is_saved_task", sa.Boolean(), server_default=sa.false(), nullable=False))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "is_saved_task")
    # ### end Alembic commands ###
