"""add failure_reason to workflow_run_blocks

Revision ID: 5be249d8dc96
Revises: f81d59b4aed5
Create Date: 2024-12-20 16:37:55.955910+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "5be249d8dc96"
down_revision: Union[str, None] = "f81d59b4aed5"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflow_run_blocks", sa.Column("failure_reason", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflow_run_blocks", "failure_reason")
    # ### end Alembic commands ###
